;;; guileobjc.scm - Scheme definitions for Objective-C interface
;   Copyright (C) 1995 R. Andrew McCallum
;
;   Written by:  R. Andrew McCallum <mccallum@gnu.ai.mit.edu>
;   Date: April 1995
;
;   This file is part of the GUILE/Objective-C Library.
;
;   This library is free software; you can redistribute it and/or
;   modify it under the terms of the GNU Library General Public
;   License as published by the Free Software Foundation; either
;   version 2 of the License, or (at your option) any later version.
;   
;   This library is distributed in the hope that it will be useful,
;   but WITHOUT ANY WARRANTY; without even the implied warranty of
;   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;   Library General Public License for more details.
;
;   You should have received a copy of the GNU Library General Public
;   License along with this library; if not, write to the Free
;   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

; Modified by: Masatake YAMATO (masata-y@is.aist-nara.ac.jp)
;	       Eiichi Takamori <taka@ma1.seikyou.ne.jp>

;;; if the executable wasn't linked against libguileobjc,
;;; then load it dynamically.  dirty hack  -eiichi ;)
(if (defined? 'objc-get-nil)
    (define-module (languages guileobjc))
  (begin
    (use-modules (languages guileobjc gocore))	; into the global namespace
    (define-module (languages guileobjc)
      :use-module (languages guileobjc gocore))))

;;; The following primitive functions defined in guileobjc.m
;;;
;;; (objc-lookup-class "ClassName")
;;; (objc-id? any-scheme-object)
;;; (objc-classnames)
;;; (objc-msg-send receiver "methodname:" arg1)

;;; Masatake adds:
;;; (objc-debug bool)
;;; (objc-new0 classname-string "init_methodname:" args ...)
	
;;; These are macro versions that require less quoting of arguments.

;;; With this you can say (objc-class NSObject) 
;;; instead of (objc-lookup-class "NSObject")

(defmacro-public objc-class (classname)
  `(objc-lookup-class (if (string? ',classname)
			  ',classname
			  (symbol->string ',classname))))

;; @protocol(Foo)
(defmacro-public objc-protocol (protocolname)
  `(objc-lookup-protocol (if (string? ',protocolname)
			  ',protocolname
			  (symbol->string ',protocolname))))

;;; With this you can say (objc-send foo respondsTo: 'myMethod:)
;;; instead of (objc-msg-send foo "respondsTo:" 'myMethod).

;; This is Andrew's original version. 
;; Spread out method notation was not supported.
(defmacro-public objc-send0 (receiver method . args)
  `(objc-msg-send ,receiver ',method ,@args))

;; This is Masatake's modifed version.
;; Spread out method notation is supported!
(defmacro-public objc-send1 (receiver arglist)
  `(objc-msg-send ,receiver ,@(objc-gather-methods-and-args (car (cdr arglist)))))

;; Merging these two. 
(defmacro-public objc-send (receiver . arglist)
  `(if (objc-spread-out-arglist? ',arglist)
       (begin (objc-send1 ,receiver ',arglist))
       (begin (objc-send0 ,receiver ,(car arglist) ,@(cdr arglist)))))

;;; Another version of the same thing
(defmacro-public [] (receiver . arglist)
  `(if (objc-spread-out-arglist? ',arglist)
       (begin (objc-send1 ,receiver ',arglist))
       (begin (objc-send0 ,receiver ,(car arglist) ,@(cdr arglist)))))

;;
;; Don't use -alloc directly form the scheme side.
;; Use this objc-new procedure instead. 
;;
(defmacro-public objc-new1 (receiver method . args)
  `(objc-new0 ,receiver ',method ,@args))
(defmacro-public objc-new2 (receiver arglist)
  `(objc-new0 ,receiver ,@(objc-gather-methods-and-args (car (cdr arglist)))))
(defmacro-public objc-new (receiver . arglist)
  `(if (objc-spread-out-arglist? ',arglist)
       (begin (objc-new2 
	       (if (catch 'misc-error
			  (lambda () ,receiver)
			  (lambda args #f))
		   (if (string? ,receiver)
		       (objc-class ,receiver) ; string
		       ,receiver)	      ; class
		   (objc-class ,receiver))    ; symbol
	       ',arglist))
       (begin (objc-new1 
	       (if (catch 'misc-error
			  (lambda () ,receiver)
			  (lambda args #f))
		   (if (string? ,receiver) 
		       (objc-class ,receiver) ; string
		       ,receiver)	      ; class
		   (objc-class ,receiver))    ; symbls
	       ,(car arglist) ,@(cdr arglist)))))

;; What is this nonsense?...
;;(define pre-guileobjc-read:sharp read:sharp)
;;(define (read:sharp c port)
;;  (if (eq? c #\[)
;;      (letrec ((read-one
;;		(lambda ()
;;		  (let ((foo (read port)))
;;		    (print foo)
;;		    (if (not (eq? foo ']))
;;			(read-one)))))))
;;      (pre-guileobjc-read:sharp c port)))
;;;#[self sendTo: foo]

;;
;; Helper functions 
;;

;; (objc-spread-out-arglist? '(abc:efg: 1 2)) -> #f
;; (objc-spread-out-arglist? '(abc: 1 efg: 2)) -> #t
(define-public (objc-spread-out-arglist? arglist)
  (let ((target (car arglist))
	(rvalue #f)
	(count-colon (lambda (x)
		       (let ((n 0))
			 (for-each (lambda (y)
				    (if (eqv? y #\:)
					(set! n (+ n 1))))
				  x) 
			 n))))
    (if (symbol? target)
	(set! target (symbol->string target)))
    (set! target (string->list target))
    (case (count-colon target)
      ((0)
       (set! rvalue #f))
      ((1)
       (if (> (length arglist) 2)
	   (begin  (set! rvalue #t))
	   (begin  (set! rvalue #f)))
       ))
    rvalue))

;; (perform: sel withObject: foo withObject: bar)
;;                         V
;; ("perform:withObject:withObject:" sel  foo  bar)
(define (objc-gather-methods-and-args arglist)
  (let ((methods '())
	(args '())
	(flag 'methods))
    (for-each (lambda (x)
		(if (eq? flag 'methods)
		    (begin 
		      (if (symbol? x)
			  (set! x (symbol->string x)))
		      (set! methods (cons x methods))
		      (set! flag 'args))
		    (begin
		      (set! args (cons x args))
		      (set! flag 'methods))))
	      arglist)
    (set! methods (reverse methods))
    (set! args (reverse args))
    (cons (list->string methods) args)))

;;
;; Bool handler(#f, #t <-> YES, NO)
;;
;; In <objc/objc-api.h>, YES is defined 1, NO is define as 0.
;;
(define-public (objc-bool->bool objcb)
  (if (eq? objcb 0)
      #f
      #t))
(define-public (bool->objc-bool scmb)
  (if scmb
      1
      0))
(define-public (objc-bool obj)
  (if (boolean? obj) 
      (bool->objc-bool obj)
      (objc-bool->bool obj)))

(define-public (objc-bool? obj)
  (or (eq? obj 1)			; ???
      (eq? obj 0)))			
;;
;; NSString handler
;;
(define-public (objc-nsstring->string objstr)
  ([] objstr cString))
(define-public (string->objc-nsstring scmstr)
  (objc-new NSString initWithCString: scmstr))
(define-public (objc-nsstring obj)
  (if (string? obj)
      (string->objc-nsstring obj)
      (objc-nsstring->string obj)))
(define-public (objc-nsstring? obj)
  (and (objc-id? obj)
       (objc-msg-send obj "isKindOf:" (objc-lookup-class "NSString"))))

;; Stolen from ScriptKit
(define-public (@ str)
  (objc-nsstring str))

;;
;; guilobjc is still buggy...e.g. on Linux a method cannot return
;; BOOL value. It always returns 0(NO). 
;; guileobjc-buggy-bool-return is a self check function.
;; If it returns #f, there is no problem to return BOOL value.
;; But if it returns #t, there is the problem.
;;
;; You can use this function at runtime.
;;
(define-public (guileobjc-buggy-bool-return)
  'unset)

(define (guileobjc-buggy-bool-return-result)
  (let* ((aobj (objc-new NSObject init))
	 (result (if (eq? (objc-bool ([] aobj isEqual: aobj)) 0)
		     #t
		     #f)))
    ;;
    (set! guileobjc-buggy-bool-return (lambda () result))
    ;;
    ))
(guileobjc-buggy-bool-return-result)

